MODULE mod_newton
  USE mod_types
  USE mod_common_var
    
  IMPLICIT NONE

CONTAINS

  SUBROUTINE sub_newton(fcn,dfcn,xguess,x,minx_in,errrel_in,errabs_in,max_iter_in,fnorm)
    ! Find zeros of function fcn, given derivative dfcn and initial condition 
    ! xguess using the basic Newton method.
    ! If the USER specificies the optional argument minx_in, only the region to
    ! the right of minx_in is searched.
    !
    ! Compulsory arguments:
    !	Inputs: 
    !		fcn: the name of the subroutine containing the function
    !           dfcn: the name of the subrouting containing the
    !		      derivative of the function
    !		xguess: initial condition for the function argument 
    !   Output:
    !		x: solution
    !
    ! Optional arguments:
    ! 	minx_in: 	lower bound on solution x 
    ! 	errrerl_in:     convergence criterion on the argument
    ! 	errabs_in:	convergence criterion on the function

    EXTERNAL :: fcn, dfcn
    REAL(dp),INTENT(in)                   :: xguess(1)
    REAL(dp), INTENT(out)                 :: x(1)
    REAL(dp),INTENT(in), OPTIONAL         :: errrel_in, errabs_in, minx_in
    INTEGER, intent(in), OPTIONAL         :: max_iter_in
    REAL(dp), INTENT(out), OPTIONAL       :: fnorm
    REAL(dp) :: dx, x_prime(1), errrel, errabs, minx, zero(1), df(1)
    INTEGER  :: iter, max_iter

    IF (PRESENT(minx_in)) THEN
      minx = minx_in
    ELSE
      minx = -10.d14
    END IF

    IF (PRESENT(errrel_in)) THEN
      errrel = errrel_in
    ELSE
      errrel = 1.d-8
    END IF

    IF (PRESENT(errabs_in)) THEN
      errabs = errabs_in
    ELSE
      errabs = 1.d-8 
    END IF

    IF (PRESENT(max_iter_in)) THEN
      max_iter = max_iter_in
    ELSE
      max_iter = 400
    END IF

    x = xguess
    DO iter = 1, max_iter
      CALL fcn(1,x,zero)
      CALL dfcn(1,x,df)

      dx = zero(1)/df(1)
      x_prime = x - dx

      IF (x_prime(1).LE.minx) THEN
        ! Ensure the algorithm stays right of lower bound on x
        x_prime = minx + 1.d-2
      END IF

      IF (ABS(df(1)) .LE. 1.d-20) THEN 
        PRINT *, "Zero derivative"
        STOP
      END IF

      IF (ABS(dx) .LE. errrel*(1+ABS(x_prime(1)))) THEN
        x = x_prime
        IF (ABS(zero(1))<errabs) THEN
          IF (PRESENT(fnorm)) fnorm = zero(1)
          RETURN
        ENDIF
      END IF
      x = x_prime
      IF (iter==201) THEN
        PRINT *,"Exceeded 200 iterations"
        PRINT *, "c_guess", x, "euler error", zero(1), &
            "h=", h_prime, "b=",b_global
        x=minx + 1.d-3
      END IF

    END DO

    PRINT *, "Exceeded maximum number of iterations"
    STOP

  END SUBROUTINE sub_newton
END MODULE mod_newton
