SUBROUTINE &
     sub_globalegm(exp_val,exp_dval,z_node,c_node,val_node,k)
! Modified egm algorithm. Solves for consumption, value fn,
!       endog. cash-at-hand nodes and their number
! Inputs: 
! 	exp_val  (DISCOUNTED expected value fn as a function of wealth ONLY )
!	exp_dval (partial derivative of exp_val with respect to wealth)
! Outputs:
!	c_node, val_node (consumption and value fn on the endogenous grid) 
!	z_node (associated endogenous grid point for cash-at-hand)
!	k (number of nodes for the policy and value fn)
!
! USER: the algorithm uses the non-linear solver in sub_newton to allow 
!       for non-invertible marginal felicity.
!       If your marginal felicity function is invertible you may replace 
!       the lines 
!             b_global = exp_dval(index_a)
!             CALL sub_newton(sub_FCN1,sub_D2FCN,xguess,x)
!       with 
!             x = u'^(-1)(exp_dval(index_a))
!       where u'^(-1) is the inverse of the marginal felicity fn.
!       In my experience, this makes hardly any difference in terms 
!       of speed.
 
  USE mod_types
  USE mod_common_var  
  USE mod_felicity  
  USE mod_fcn        
  USE mod_newton
!  USE, intrinsic:: ieee_arithmetic

  IMPLICIT NONE

  REAL (dp), INTENT(in),  DIMENSION(len_a_grid)	:: exp_val, exp_dval
  REAL (dp), INTENT(out), DIMENSION(len_a_grid) :: c_node, val_node, z_node
  INTEGER, INTENT(out)				:: k

  INTEGER  :: i, index_a, index_a_min, index_a_max, j, j_0, j_bar, max_loc
  INTEGER,   DIMENSION(len_a_grid) :: aprime_node

  REAL(dp) :: bound, c_bar, exp_dval_down, exp_dval_up, max_value, &
       c_t(1), v(1), v_bar, z_bar, z_temp, indic, val_temp
  REAL(dp), ALLOCATABLE :: c_temp(:)
  REAL(dp), ALLOCATABLE :: euler(:), val(:)

  LOGICAL, DIMENSION(SIZE(exp_val)-1)     :: mask
  
  !Variables used by sub_newton subroutine
  INTEGER, PARAMETER    :: N=1
  REAL(dp)              :: FNORM, F(N), X(N), XGUESS(N)
    

  ! Initialize indices of the bounds of the concave region
 index_a_min = len_a_grid
 index_a_max = 1

  !1. Check for non-concavity and compute bounds of non-concave region
  !-------------------------------------------------------------------
  ! Find discontinuities (upward jumps) in future marginal utility
  mask = (exp_dval(2:len_a_grid)>exp_dval(1:len_a_grid-1))
  IF (COUNT(mask)>0) THEN 
     ! Non-concavity. Computes bounds of region.
     exp_dval_up = MAXVAL(PACK(exp_dval(2:len_a_grid),mask))
     exp_dval_down = MINVAL(PACK(exp_dval(1:len_a_grid-1),mask))
     IF (COUNT(exp_dval>exp_dval_up)==0) THEN 
        index_a_min=1
     ELSE
        index_a_min = MINLOC(exp_dval,1,exp_dval>exp_dval_up)+1
     END IF
     IF (COUNT(exp_dval<exp_dval_down)==0) THEN 
        index_a_max=len_a_grid 
     ELSE
        index_a_max = MAXLOC(exp_dval,1,exp_dval<exp_dval_down) - 1
     END IF
  END IF

  !2. Solve for endogenous grid points
  !------------------------------------------------------------------
  ! Initializing parameters for the egm algorithm
  
  xguess = 1.d-1
  k = 0
  j_0 = 1 
  j_bar = 0
  z_bar = 10.d6
  v_bar = exp_dval_up + 1.d-5

  ! Loop over future asset levels
  DO index_a=1,len_a_grid

    IF (index_a<index_a_min .OR. index_a>index_a_max) THEN  !if_1
      ! 2.1 Concave region - standard egm 
      ! ----------------------------------
      b_global = exp_dval(index_a)
      CALL sub_newton(sub_FCN1,sub_D2FCN,xguess,x,0.d0)
      k = k+1
      c_node(k) = x(1)
      z_node(k) = c_node(k) + a_grid(index_a)
      aprime_node(k) = index_a 
      xguess = x
    ELSE
      !2.2 Non-concave region 
      ! ----------------------------------
      
      !2.2.1 If the policy function has been solved for at least one
      !      point update lower bounds used to exploit monotonicity
      IF (k>0) THEN 
        c_bar =MAX(z_node(k)-a_grid(index_a),1.d-5)
        v_bar=dfel(c_bar)
        j_0 = aprime_node(k) 
      END IF
      bound = dfel(z_bar-a_grid(index_a))

      ! Check that bounds are real (used only to debug code)
      !        IF (ieee_is_nan(v_bar) .OR. ieee_is_nan(bound)) THEN
      !            print *, "NAN"
      !        endif   

      !2.2.2 Find local max exploiting monotonicity
      IF (  (v_bar>exp_dval(index_a) .AND.  exp_dval(index_a).GE.bound) .OR. &
          (exp_dval(index_a)<bound .AND. index_a.GE.j_bar)  ) THEN !if_2
        ! Use standard egm to find LOCAL max
        b_global = exp_dval(index_a)
        CALL sub_newton(sub_fcn1,sub_d2fcn,xguess,x,0.d0)
        z_temp = x(1) + a_grid(index_a)

        ! Verify that a_grid(index_a) is a GLOBAL max given z_temp
        ALLOCATE(c_temp(j_0:index_a_max))
        DO j = j_0,index_a_max
          c_temp(j) = z_temp - a_grid(j)
          IF (c_temp(j).LE.0 .AND. j>1) EXIT
        END DO
        ALLOCATE (val(j_0:j-1))
        val =  fel(c_temp(j_0:j-1))+ exp_val(j_0:j-1)
        max_loc = MAXLOC(val,1)+j_0-1
        DEALLOCATE (c_temp,val)
        IF (max_loc==index_a) THEN            !if_3
          ! a_grid(index_a) is GLOBAL max given z_temp
          ! Store policy fns and associated cash at hand 
          k = k+1
          c_node(k) = x(1)
          z_node(k) = z_temp
          aprime_node(k) = index_a
          xguess = x
          z_bar = 10.d6
        ELSEIF (max_loc>index_a) THEN 
          ! Store location of (approximate) global max given z_temp
          ! By monotonicity no need to consider points to its left.
          j_bar=max_loc
          z_bar = z_temp
        END IF   !if_3
      END IF     !if_2
    END IF       !if_1
  END DO

  !3. Solve for value fn on the endog. cash-at-hand nodes
  !----------------------------------------------------------

  val_node(1:k) = fel(c_node(1:k)) + exp_val(aprime_node(1:k))

  !4. Correct the policy and value fns on the first end. cash-at-hand
  !   node if the choice of future assets is to the right of the
  !   borrowing constraint
  !----------------------------------------------------------

  IF (aprime_node(1)>1) THEN
    b_global = a_grid(aprime_node(1))
    fut_v1_global = exp_val(aprime_node(1)) - exp_val(1)
    xguess = z_node(1)
    ! Solve for value of cash at hand to the left of which it is 
    ! optimal to be borrowing constrained and for the associated values
    ! of the consumption and value fn
    CALL sub_newton(sub_min_z,sub_dmin_z,xguess,x,b_global)
    c_node(2:len_a_grid) = c_node(1:len_a_grid-1)
    z_node(2:len_a_grid) = z_node(1:len_a_grid-1)
    val_node(2:len_a_grid) = val_node(1:len_a_grid-1)
    c_node(1) = -1.d-5 + x(1) 
    z_node(1) = c_node(1) + a_grid(1)
    val_temp = &
        fel(c_node(1))+ exp_val(1)
    val_node(1) = val_temp
    k = k + 1 
  END IF

END SUBROUTINE sub_globalegm
